import { _decorator, Component, Node, AudioSource, AudioClip, Vec3 } from "cc";
const { ccclass, property } = _decorator;

@ccclass("audio3D")
export class audio3D extends Component {
  @property({
    type: AudioClip,
    displayName: "音频源",
  })
  audioClip: AudioClip;

  @property({
    type: Node,
    displayName: "发声者",
  })
  pannerNode: Node;

  @property({
    type: Node,
    displayName: "收听者",
  })
  listenerNode: Node;

  @property({
    type: Number,
    displayName: "最大距离",
  })
  maxDistance: number = 100;

  @property({
    type: Boolean,
    displayName: "是否循环播放",
  })
  loop: boolean = true;

  @property({
    type: Number,
    displayName: "音量",
  })
  volume: number = 1;

  @property({
    type: Number,
    displayName: "播放速度",
  })
  playbackRate: number = 1;

  @property({
    type: Boolean,
    displayName: "自动播放",
  })
  playOnAwake: boolean = true;


  panner: PannerNode;
  listener: AudioListener;
  source: AudioBufferSourceNode;
  gainNode: GainNode;

  ac: AudioContext;

  start() {
    this.ac = new AudioContext();
    this.init();
  }

  async init() {

    this.source = this.ac.createBufferSource();
    // @ts-ignore
    this.source.buffer = this.audioClip._player._player._audioBuffer;

    
    // source.start();
    this.source.loop = this.loop;

    this.gainNode = this.ac.createGain();
    this.panner = this.ac.createPanner();
    this.panner.panningModel = "equalpower"; // 音频空间化算法模型
    this.panner.distanceModel = "linear"; // 远离时的音量衰减算法
    this.panner.maxDistance = this.maxDistance; // 最大距离
    this.panner.refDistance = 5; // 开始衰减的参考距离
    this.panner.rolloffFactor = 3; // 衰减速度
    this.panner.coneInnerAngle = 360; // 声音360度扩散

    this.panner.orientationX.value = 1; // 声源朝向x分量
    this.panner.orientationY.value = 0;
    this.panner.orientationZ.value = 0;

    this.listener = this.ac.listener;

    this.gainNode.connect(this.ac.destination);
    this.gainNode.connect(this.panner);

    this.source.connect(this.panner);
    this.panner.connect(this.gainNode);

    console.log("this.source",this.source,this.gainNode,this.ac)

    this.source.playbackRate.value = Math.floor(this.playbackRate * 100) / 100;
    let _volume = Math.floor(this.volume * 100) / 100;
    this.gainNode.gain.value = _volume;

    if(this.playOnAwake){
        this.source.start();
    }
  }

  _distance = 0;

  update(deltaTime: number) {
    if (!this.pannerNode) {
      return;
    }

    if (!this.listenerNode) {
      return;
    }

    this.panner.setPosition(
      this.pannerNode.worldPosition.x,
      this.pannerNode.worldPosition.y,
      this.pannerNode.worldPosition.z
    );

    let _worldRotation = new Vec3(0, 0, 0);
    this.listenerNode.worldRotation.getEulerAngles(_worldRotation);

    this.listener.setOrientation(
      this.listenerNode.forward.x || 0,
      this.listenerNode.forward.y || 0,
      this.listenerNode.forward.z || 0,
      this.listenerNode.up.x || 0,
      this.listenerNode.up.y || 1,
      this.listenerNode.up.z || 0
    );
    this.listener.setPosition(
      this.listenerNode.worldPosition.x || 0,
      this.listenerNode.worldPosition.y || 0,
      this.listenerNode.worldPosition.z || 0
    );

    this.source.playbackRate.value = Math.floor(this.playbackRate * 100) / 100;
    let _volume = Math.floor(this.volume * 100) / 100;
    this.gainNode.gain.value = _volume;
  }

  onDestroy() {
    try {
      this.gainNode.disconnect(this.ac.destination);
      this.gainNode.disconnect(this.panner);

      this.source.disconnect(this.panner);
      this.panner.disconnect(this.gainNode);

      this.source.stop();
      this.ac.close();
    } catch(e) {
        console.log(e);
    }
  }
}
